/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include <string.h>
#include <sys/time.h>

#include "aasink.h"

static GstElementDetails gst_aasink_details = {
  "Video sink",
  "Sink/Video",
  "An ASCII art videosink",
  VERSION,
  "Wim Taymans <wim.taymans@chello.be>",
  "(C) 2001",
};

/* aasink signals and args */
enum {
  SIGNAL_FRAME_DISPLAYED,
  SIGNAL_HAVE_SIZE,
  LAST_SIGNAL
};


enum {
  ARG_0,
  ARG_WIDTH,
  ARG_HEIGHT,
  ARG_DRIVER,
  ARG_DITHER,
  ARG_BRIGHTNESS,
  ARG_CONTRAST,
  ARG_GAMMA,
  ARG_INVERSION,
  ARG_RANDOMVAL,
  ARG_FRAMES_DISPLAYED,
  ARG_FRAME_TIME,
};

GST_PADTEMPLATE_FACTORY (sink_template,
  "sink",
  GST_PAD_SINK,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW (
    "aasink_caps",
    "video/raw",
      "format", 	GST_PROPS_FOURCC (GST_STR_FOURCC ("I420"))
  )
)
    
static void	gst_aasink_class_init	(GstAASinkClass *klass);
static void	gst_aasink_init		(GstAASink *aasink);

static void	gst_aasink_chain		(GstPad *pad, GstBuffer *buf);

static void	gst_aasink_set_arg		(GtkObject *object, GtkArg *arg, guint id);
static void	gst_aasink_get_arg		(GtkObject *object, GtkArg *arg, guint id);

static GstElementStateReturn gst_aasink_change_state (GstElement *element);

static GstElementClass *parent_class = NULL;
static guint gst_aasink_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_aasink_get_type (void)
{
  static GtkType aasink_type = 0;

  if (!aasink_type) {
    static const GtkTypeInfo aasink_info = {
      "GstAASink",
      sizeof(GstAASink),
      sizeof(GstAASinkClass),
      (GtkClassInitFunc)gst_aasink_class_init,
      (GtkObjectInitFunc)gst_aasink_init,
      (GtkArgSetFunc)gst_aasink_set_arg,
      (GtkArgGetFunc)gst_aasink_get_arg,
      (GtkClassInitFunc)NULL,
    };
    aasink_type = gtk_type_unique(GST_TYPE_ELEMENT,&aasink_info);
  }
  return aasink_type;
}

#define GST_TYPE_AADRIVERS (gst_aasink_drivers_get_type())
static GtkType
gst_aasink_drivers_get_type (void)
{
  static GtkType driver_type = 0;
  if (!driver_type) {
    GtkEnumValue *drivers;
    struct aa_driver *driver;
    gint i = 0;

    driver = aa_drivers[i++];
    while (driver) {
      driver = aa_drivers[i++];
    }
    
    drivers = g_new0(GtkEnumValue, i);

    i = 0;
    driver = aa_drivers[i];
    while (driver) {
      drivers[i].value = i;
      drivers[i].value_name = g_strdup (driver->shortname);
      drivers[i].value_nick = g_strdup (driver->name);
      i++;
      driver = aa_drivers[i];
    }
    drivers[i].value = 0;
    drivers[i].value_name = NULL;
    drivers[i].value_nick = NULL;

    driver_type = gtk_type_register_enum ("GstAASinkDrivers", drivers);
  }
  return driver_type;
}

#define GST_TYPE_AADITHER (gst_aasink_dither_get_type())
static GtkType
gst_aasink_dither_get_type (void)
{
  static GtkType dither_type = 0;
  if (!dither_type) {
    GtkEnumValue *ditherers;
    gint i = 0;

    while (aa_dithernames[i]) {
      i++;
    }
    
    ditherers = g_new0(GtkEnumValue, i);

    i = 0;
    while (aa_dithernames[i]) {
      ditherers[i].value = i;
      ditherers[i].value_name = g_strdup (aa_dithernames[i]);
      ditherers[i].value_nick = g_strdup (aa_dithernames[i]);
      i++;
    }
    ditherers[i].value = 0;
    ditherers[i].value_name = NULL;
    ditherers[i].value_nick = NULL;

    dither_type = gtk_type_register_enum ("GstAASinkDitherers", ditherers);
  }
  return dither_type;
}

static void
gst_aasink_class_init (GstAASinkClass *klass)
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class (GST_TYPE_ELEMENT);

  gtk_object_add_arg_type ("GstAASink::width", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_WIDTH);
  gtk_object_add_arg_type ("GstAASink::height", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_HEIGHT);
  gtk_object_add_arg_type ("GstAASink::driver", GST_TYPE_AADRIVERS,
                           GTK_ARG_READWRITE, ARG_DRIVER);
  gtk_object_add_arg_type ("GstAASink::dither", GST_TYPE_AADITHER,
                           GTK_ARG_READWRITE, ARG_DITHER);
  gtk_object_add_arg_type ("GstAASink::brightness", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_BRIGHTNESS);
  gtk_object_add_arg_type ("GstAASink::contrast", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_CONTRAST);
  gtk_object_add_arg_type ("GstAASink::gamma", GTK_TYPE_FLOAT,
                           GTK_ARG_READWRITE, ARG_GAMMA);
  gtk_object_add_arg_type ("GstAASink::inversion", GTK_TYPE_BOOL,
                           GTK_ARG_READWRITE, ARG_INVERSION);
  gtk_object_add_arg_type ("GstAASink::randomval", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_RANDOMVAL);
  gtk_object_add_arg_type ("GstAASink::frames_displayed", GTK_TYPE_INT,
                           GTK_ARG_READABLE, ARG_FRAMES_DISPLAYED);
  gtk_object_add_arg_type ("GstAASink::frame_time", GTK_TYPE_INT,
                           GTK_ARG_READABLE, ARG_FRAME_TIME);

  gtkobject_class->set_arg = gst_aasink_set_arg;
  gtkobject_class->get_arg = gst_aasink_get_arg;

  gst_aasink_signals[SIGNAL_FRAME_DISPLAYED] =
    gtk_signal_new ("frame_displayed", GTK_RUN_LAST, gtkobject_class->type,
                    GTK_SIGNAL_OFFSET (GstAASinkClass, frame_displayed),
                    gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  gst_aasink_signals[SIGNAL_HAVE_SIZE] =
    gtk_signal_new ("have_size", GTK_RUN_LAST, gtkobject_class->type,
                    GTK_SIGNAL_OFFSET (GstAASinkClass, have_size),
                    gtk_marshal_NONE__INT_INT, GTK_TYPE_NONE, 2,
		    GTK_TYPE_UINT, GTK_TYPE_UINT);

  gtk_object_class_add_signals (gtkobject_class, gst_aasink_signals,
                                  LAST_SIGNAL);

  gstelement_class->change_state = gst_aasink_change_state;
}

static void
gst_aasink_newcaps (GstPad *pad, GstCaps *caps)
{
  GstAASink *aasink;
  gulong print_format;

  aasink = GST_AASINK (gst_pad_get_parent (pad));

  aasink->width =  gst_caps_get_int (caps, "width");
  aasink->height =  gst_caps_get_int (caps, "height");

  print_format = GULONG_FROM_LE (aasink->format);

  GST_DEBUG (0, "aasink: setting %08lx (%4.4s)\n", aasink->format, (gchar*)&print_format);
  
  gtk_signal_emit (GTK_OBJECT (aasink), gst_aasink_signals[SIGNAL_HAVE_SIZE], 
		  aasink->width, aasink->height);
}

static void
gst_aasink_init (GstAASink *aasink)
{
  aasink->sinkpad = gst_pad_new_from_template (
		  GST_PADTEMPLATE_GET (sink_template), "sink");
  gst_element_add_pad (GST_ELEMENT (aasink), aasink->sinkpad);
  gst_pad_set_chain_function (aasink->sinkpad, gst_aasink_chain);
  gst_pad_set_newcaps_function (aasink->sinkpad, gst_aasink_newcaps);

  aasink->clock = gst_clock_get_system();
  gst_clock_register(aasink->clock, GST_OBJECT(aasink));

  memcpy(&aasink->ascii_surf, &aa_defparams, sizeof (struct aa_hardware_params));
  aasink->ascii_parms.bright = 0;
  aasink->ascii_parms.contrast = 16;
  aasink->ascii_parms.gamma = 1.0;
  aasink->ascii_parms.dither = 0;
  aasink->ascii_parms.inversion = 0;
  aasink->ascii_parms.randomval = 0;
  aasink->aa_driver = 0;

  aasink->width = -1;
  aasink->height = -1;

  GST_FLAG_SET(aasink, GST_ELEMENT_THREAD_SUGGESTED);
}

static void
gst_aasink_scale (GstAASink *aasink, gchar *src, gchar *dest,
		  gint sw, gint sh, 
		  gint dw, gint dh)
{
  gint ypos, yinc, y;
  gint xpos, xinc, x;

  g_return_if_fail ((dw != 0) && (dh != 0));

  ypos = 0x10000;
  yinc = (sh << 16) / dh;
  xinc = (sw << 16) / dw;

  for (y = dh; y; y--) {
    while (ypos > 0x10000) {
      ypos -= 0x10000;
      src  += sw;
    }
    xpos = 0x10000;
    {
      guchar *destp = dest;
      guchar *srcp = src;

      for ( x=dw; x; x-- ) {
        while (xpos >= 0x10000L) {
          srcp++;
          xpos -= 0x10000L;
        }
        *destp++ = *srcp;
        xpos     += xinc;
      }
    }
    dest += dw;
    ypos += yinc;
  }
}

static void
gst_aasink_chain (GstPad *pad, GstBuffer *buf)
{
  GstAASink *aasink;
  GstClockTimeDiff jitter;

  g_return_if_fail (pad != NULL);
  g_return_if_fail (GST_IS_PAD (pad));
  g_return_if_fail (buf != NULL);

  aasink = GST_AASINK (gst_pad_get_parent (pad));

  gst_aasink_scale (aasink,
		    GST_BUFFER_DATA (buf),		/* src */
  		    aa_image (aasink->context),		/* dest */
  		    aasink->width,			/* sw */
  		    aasink->height,			/* sh */
  		    aa_imgwidth (aasink->context),	/* dw */
  		    aa_imgheight (aasink->context));	/* dh */

  if (!GST_BUFFER_FLAG_IS_SET(buf, GST_BUFFER_FLUSH)) {
    GST_DEBUG (0,"videosink: clock wait: %llu\n", GST_BUFFER_TIMESTAMP(buf));

    jitter = gst_clock_current_diff(aasink->clock, GST_BUFFER_TIMESTAMP (buf));

    if (jitter > 500000 || jitter < -500000)
    {
      GST_DEBUG (0, "jitter: %lld\n", jitter);
      gst_clock_set (aasink->clock, GST_BUFFER_TIMESTAMP (buf));
    }
    else {
      gst_clock_wait(aasink->clock, GST_BUFFER_TIMESTAMP(buf), GST_OBJECT(aasink));
    }
  }

  aa_render (aasink->context, &aasink->ascii_parms, 
		  0, 0, aa_imgwidth (aasink->context), aa_imgheight (aasink->context));
  aa_flush (aasink->context);
  aa_getevent (aasink->context, FALSE);

  gtk_signal_emit(GTK_OBJECT(aasink),gst_aasink_signals[SIGNAL_FRAME_DISPLAYED]);

  gst_buffer_unref(buf);
}


static void
gst_aasink_set_arg (GtkObject *object, GtkArg *arg, guint id)
{
  GstAASink *aasink;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail (GST_IS_AASINK (object));

  aasink = GST_AASINK (object);

  switch(id) {
    case ARG_WIDTH:
      aasink->ascii_surf.width = GTK_VALUE_INT(*arg);
      break;
    case ARG_HEIGHT:
      aasink->ascii_surf.height = GTK_VALUE_INT(*arg);
      break;
    case ARG_DRIVER: {
      aasink->aa_driver = GTK_VALUE_INT(*arg);
      break;
    }
    case ARG_DITHER: {
      aasink->ascii_parms.dither = GTK_VALUE_INT (*arg);
      break;
    }
    case ARG_BRIGHTNESS: {
      aasink->ascii_parms.bright = GTK_VALUE_INT (*arg);
      break;
    }
    case ARG_CONTRAST: {
      aasink->ascii_parms.contrast = GTK_VALUE_INT (*arg);
      break;
    }
    case ARG_GAMMA: {
      aasink->ascii_parms.gamma = GTK_VALUE_FLOAT (*arg);
      break;
    }
    case ARG_INVERSION: {
      aasink->ascii_parms.inversion = GTK_VALUE_BOOL (*arg);
      break;
    }
    case ARG_RANDOMVAL: {
      aasink->ascii_parms.randomval = GTK_VALUE_INT (*arg);
      break;
    }
    default:
      break;
  }
}

static void
gst_aasink_get_arg (GtkObject *object, GtkArg *arg, guint id)
{
  GstAASink *aasink;

  /* it's not null if we got it, but it might not be ours */
  aasink = GST_AASINK(object);

  switch(id) {
    case ARG_WIDTH: {
      GTK_VALUE_INT(*arg) = aasink->ascii_surf.width;
      break;
    }
    case ARG_HEIGHT: {
      GTK_VALUE_INT(*arg) = aasink->ascii_surf.height;
      break;
    }
    case ARG_DRIVER: {
      GTK_VALUE_INT(*arg) = aasink->aa_driver;
      break;
    }
    case ARG_DITHER: {
      GTK_VALUE_INT(*arg) = aasink->ascii_parms.dither;
      break;
    }
    case ARG_BRIGHTNESS: {
      GTK_VALUE_INT(*arg) = aasink->ascii_parms.bright;
      break;
    }
    case ARG_CONTRAST: {
      GTK_VALUE_INT(*arg) = aasink->ascii_parms.contrast;
      break;
    }
    case ARG_GAMMA: {
      GTK_VALUE_FLOAT(*arg) = aasink->ascii_parms.gamma;
      break;
    }
    case ARG_INVERSION: {
      GTK_VALUE_INT(*arg) = aasink->ascii_parms.inversion;
      break;
    }
    case ARG_RANDOMVAL: {
      GTK_VALUE_BOOL(*arg) = aasink->ascii_parms.randomval;
      break;
    }
    case ARG_FRAMES_DISPLAYED: {
      GTK_VALUE_INT(*arg) = aasink->frames_displayed;
      break;
    }
    case ARG_FRAME_TIME: {
      GTK_VALUE_INT(*arg) = aasink->frame_time/1000000;
      break;
    }
    default: {
      arg->type = GTK_TYPE_INVALID;
      break;
    }
  }
}

static gboolean
gst_aasink_open (GstAASink *aasink)
{
  g_return_val_if_fail (!GST_FLAG_IS_SET (aasink ,GST_AASINK_OPEN), FALSE);

  aa_recommendhidisplay(aa_drivers[aasink->aa_driver]->shortname);

  aasink->context = aa_autoinit (&aasink->ascii_surf);
  if (aasink->context == NULL) {
    gst_element_error (GST_ELEMENT (aasink), 
		    g_strdup("opening aalib context"));
    return FALSE;
  }
  aa_autoinitkbd(aasink->context, 0);
  aa_resizehandler(aasink->context, (void *)aa_resize);

  GST_FLAG_SET (aasink, GST_AASINK_OPEN);

  return TRUE;
}

static void
gst_aasink_close (GstAASink *aasink)
{
  g_return_if_fail (GST_FLAG_IS_SET (aasink ,GST_AASINK_OPEN));

  aa_close (aasink->context);

  GST_FLAG_UNSET (aasink, GST_AASINK_OPEN);
}

static GstElementStateReturn
gst_aasink_change_state (GstElement *element)
{
  g_return_val_if_fail (GST_IS_AASINK (element), GST_STATE_FAILURE);

  if (GST_STATE_PENDING (element) == GST_STATE_NULL) {
    if (GST_FLAG_IS_SET (element, GST_AASINK_OPEN))
      gst_aasink_close (GST_AASINK (element));
  } else {
    if (!GST_FLAG_IS_SET (element, GST_AASINK_OPEN)) {
      if (!gst_aasink_open (GST_AASINK (element)))
        return GST_STATE_FAILURE;
    }
  }

  if (GST_ELEMENT_CLASS (parent_class)->change_state)
    return GST_ELEMENT_CLASS (parent_class)->change_state (element);

  return GST_STATE_SUCCESS;
}

static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  /* create an elementfactory for the aasink element */
  factory = gst_elementfactory_new("aasink",GST_TYPE_AASINK,
                                   &gst_aasink_details);
  g_return_val_if_fail(factory != NULL, FALSE);

  gst_elementfactory_add_padtemplate (factory, 
		  GST_PADTEMPLATE_GET (sink_template));

  gst_plugin_add_factory(plugin,factory);

  return TRUE;
}

GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "aasink",
  plugin_init
};
