/* Gnome-Streamer
 * Copyright (C) <2001> Steve Baker <stevebaker_org@yahoo.co.uk>
 *
 * float2int.c
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <float2int.h>


static GstElementDetails float2int_details = {
  "Float to Integer effect",
  "Filter/Effect",
  "Convert from floating point to integer audio data",
  VERSION,
  "Steve Baker <stevebaker_org@yahoo.co.uk>",
  "(C) 2001",
};


enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_PAN,
};

static GstPadTemplate* 
float2int_sink_factory (void) 
{
  return 
    gst_padtemplate_new (
      "sink",
      GST_PAD_SINK,
      GST_PAD_ALWAYS,
      gst_caps_new (
      "float2int_sink",
        "audio/raw",
        gst_props_new (
          "format",     GST_PROPS_STRING ("float"),
          "layout",     GST_PROPS_STRING ("gfloat"),
          "intercept",  GST_PROPS_FLOAT(0.0),
          "slope",      GST_PROPS_FLOAT(1.0),
          "channels",   GST_PROPS_INT (1),
      NULL)),
    NULL);
}

static GstPadTemplate* 
float2int_src_factory (void) 
{
  return 
    gst_padtemplate_new (
    "src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    gst_caps_new (
    "float2int_src",
      "audio/raw",
      gst_props_new (
        "format",   GST_PROPS_STRING ("int"),
        "law",     GST_PROPS_INT (0),
        "endianness",     GST_PROPS_INT (G_BYTE_ORDER),
        "signed",   GST_PROPS_BOOLEAN (TRUE),
        "channels", GST_PROPS_INT (1),
        NULL)),
      NULL);
}

static GstPadTemplate *srctempl, *sinktempl;

static void	gst_float2int_class_init		(GstFloat2IntClass *klass);
static void	gst_float2int_init			(GstFloat2Int *plugin);
static void	gst_float2int_set_arg		(GtkObject *object, GtkArg *arg, guint id);
static void	gst_float2int_get_arg		(GtkObject *object, GtkArg *arg, guint id);
static void	gst_float2int_chain_gint16		(GstPad *pad, GstBuffer *buf_in);
static inline GstFloat2Int* gst_float2int_get_plugin(GstPad *pad,GstBuffer *buf);

static GstElementClass *parent_class = NULL;

static GstPadNegotiateReturn
gst_float2int_negotiate_src (GstPad *pad, GstCaps **caps, gpointer *data)
{
  GstFloat2Int* plugin = GST_FLOAT2INT (GST_OBJECT_PARENT (pad));
  GstCaps* peercaps;
  GstRealPad *realpeer;
  GstPadNegotiateReturn proxyresult;

  g_print("nego float2int src\n");
  if (*caps== NULL) return GST_PAD_NEGOTIATE_FAIL;
  plugin->width = gst_caps_get_int (*caps, "width");
  plugin->depth = gst_caps_get_int (*caps, "depth");

  if (plugin->trying_proxy_caps) return GST_PAD_NEGOTIATE_AGREE;

  plugin->trying_proxy_caps=TRUE;
  
  peercaps=gst_caps_new (
    "float2int_src_caps",
    "audio/raw",
    gst_props_new (
      "format",     GST_PROPS_STRING ("float"),
      "layout",     GST_PROPS_STRING ("gfloat"),
      "intercept",  GST_PROPS_FLOAT(0.0),
      "slope",      GST_PROPS_FLOAT(1.0),
      "rate",       GST_PROPS_INT (gst_caps_get_int (*caps, "rate")),
      "channels",   GST_PROPS_INT (1),
      NULL
    )
  );
  realpeer = GST_RPAD_PEER(plugin->sinkpad);
  proxyresult = gst_pad_negotiate_proxy(&(realpeer->pad), plugin->sinkpad, &peercaps);
  plugin->trying_proxy_caps=FALSE;
  return proxyresult;
}

static GstPadNegotiateReturn
gst_float2int_negotiate_sink (GstPad *pad, GstCaps **caps, gpointer *data)
{
  GstFloat2Int* plugin = GST_FLOAT2INT (GST_OBJECT_PARENT (pad));
  GstCaps* peercaps;
  GstRealPad *realpeer;
  GstPadNegotiateReturn proxyresult;

  g_print("nego float2int sink\n");
  if (*caps== NULL) return GST_PAD_NEGOTIATE_FAIL;  

  if (plugin->trying_proxy_caps) return GST_PAD_NEGOTIATE_AGREE;

  plugin->trying_proxy_caps=TRUE;
  peercaps=gst_caps_new (
    "float2int_sink_caps",
    "audio/raw",
    gst_props_new (
      "format",     GST_PROPS_STRING ("int"),
      "law",        GST_PROPS_INT (0),
      "endianness", GST_PROPS_INT (G_BYTE_ORDER),
      "signed",     GST_PROPS_BOOLEAN (TRUE),
      "width",      GST_PROPS_INT (plugin->width),
      "depth",      GST_PROPS_INT (plugin->depth),
      "rate",       GST_PROPS_INT (gst_caps_get_int (*caps, "rate")),
      "channels",   GST_PROPS_INT (1),
      NULL
    )
  );
  realpeer = GST_RPAD_PEER(plugin->srcpad);
  proxyresult = gst_pad_negotiate_proxy(&(realpeer->pad), plugin->srcpad, &peercaps);
  plugin->trying_proxy_caps=FALSE;
  return proxyresult;
}

GtkType
gst_float2int_get_type(void) {
  static GtkType float2int_type = 0;

  if (!float2int_type) {
    static const GtkTypeInfo float2int_info = {
      "GstFloat2Int",
      sizeof(GstFloat2Int),
      sizeof(GstFloat2IntClass),
      (GtkClassInitFunc)gst_float2int_class_init,
      (GtkObjectInitFunc)gst_float2int_init,
      (GtkArgSetFunc)gst_float2int_set_arg,
      (GtkArgGetFunc)gst_float2int_get_arg,
      (GtkClassInitFunc)NULL,
    };
    float2int_type = gtk_type_unique(GST_TYPE_ELEMENT,&float2int_info);
  }
  return float2int_type;
}

static void
gst_float2int_class_init (GstFloat2IntClass *klass)
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtkobject_class->set_arg = gst_float2int_set_arg;
  gtkobject_class->get_arg = gst_float2int_get_arg;
}

static void
gst_float2int_init (GstFloat2Int *plugin)
{
  plugin->sinkpad = gst_pad_new_from_template(sinktempl,"sink");
  plugin->srcpad = gst_pad_new_from_template(srctempl,"src");

  gst_element_add_pad(GST_ELEMENT(plugin),plugin->sinkpad);
  gst_pad_set_chain_function(plugin->sinkpad,gst_float2int_chain_gint16);
  gst_element_add_pad(GST_ELEMENT(plugin),plugin->srcpad);

  gst_pad_set_negotiate_function(plugin->sinkpad, gst_float2int_negotiate_sink);
  gst_pad_set_negotiate_function(plugin->srcpad, gst_float2int_negotiate_src);
  plugin->channels=1;
  plugin->width=16;
  plugin->depth=16;
  plugin->intercept=0.0;
  plugin->slope=1.0;
  plugin->trying_proxy_caps=FALSE;
}

static void
gst_float2int_chain_gint16 (GstPad *pad,GstBuffer *buf_in)
{
  GstFloat2Int *plugin;
  GstBuffer *buf_out;
  gfloat *data_in;
  gint16 *data_out;
  gint num_samples, i;
  
  g_return_if_fail(plugin = gst_float2int_get_plugin(pad,buf_in));  

  num_samples = GST_BUFFER_SIZE(buf_in) / sizeof(gfloat);
  data_in=(gfloat *)GST_BUFFER_DATA(buf_in);

  buf_out = gst_buffer_new();
  g_return_if_fail (buf_out);
  data_out = g_new(gint16, num_samples);
  GST_BUFFER_DATA(buf_out) = (gpointer) data_out;
  GST_BUFFER_SIZE(buf_out) = 2 * num_samples;
  
  for (i = 0; i < num_samples; i++) {
    data_out[i] = (gint16)(data_in[i] * 32767.0);
  }
  
  gst_buffer_unref(buf_in);  
  gst_pad_push(plugin->srcpad,buf_out);
}

static inline GstFloat2Int* 
gst_float2int_get_plugin(GstPad *pad,GstBuffer *buf)
{
  GstFloat2Int *plugin;
  g_return_val_if_fail(pad != NULL, NULL);
  g_return_val_if_fail(GST_IS_PAD(pad), NULL);
  g_return_val_if_fail(buf != NULL, NULL);

  plugin = GST_FLOAT2INT(GST_OBJECT_PARENT (pad));
  g_return_val_if_fail(plugin != NULL, NULL);
  g_return_val_if_fail(GST_IS_FLOAT2INT(plugin), NULL);
  return plugin;
}

static void
gst_float2int_set_arg (GtkObject *object, GtkArg *arg, guint id)
{
  GstFloat2Int *plugin;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_FLOAT2INT(object));
  plugin = GST_FLOAT2INT(object);

  switch(id) {
    default:
      break;
  }
}

static void
gst_float2int_get_arg (GtkObject *object, GtkArg *arg, guint id)
{
  GstFloat2Int *plugin;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_FLOAT2INT(object));
  plugin = GST_FLOAT2INT(object);

  switch (id) {
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}

gboolean 
gst_float2int_factory_init (GstPlugin *plugin) 
{
  GstElementFactory *factory;
  
  factory = gst_elementfactory_new("float2int",GST_TYPE_FLOAT2INT,
                                   &float2int_details);
  g_return_val_if_fail(factory != NULL, FALSE);
  
  sinktempl = float2int_sink_factory();
  gst_elementfactory_add_padtemplate (factory, sinktempl);

  srctempl = float2int_src_factory();
  gst_elementfactory_add_padtemplate (factory, srctempl);
  
  gst_plugin_add_factory(plugin,factory);
   
  return TRUE;
}