/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


//#define GST_DEBUG_ENABLED
#include <string.h>
#include <openquicktime/openquicktime.h>
#include "gstquicktimedecoder.h"

static void gst_mov_decoder_chain (GstPad *pad, GstBuffer *buf);
static void gst_mov_decoder_loop (GstElement *element);
static quicktime_t* quicktime_open_pad(GstMovDecoder *mov_decoder);
static GstCaps* mov_typefind (GstBuffer *buf, gpointer private);

/* elementfactory information */
static GstElementDetails gst_mov_decoder_details = {
  "quicktime parser",
  "Parser/Video",
  "Parse a quicktime file and decompressed it into raw audio and video",
  VERSION,
  "Yann <yann@3ivx.com>",
  "(C) 2001",
};



  /* typefactory for 'mov' */
static GstTypeFactory movfactory = {
  "video/mov",
  ".mov",
  mov_typefind,
};



/* MovDecoder signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_BIT_RATE,
  ARG_MEDIA_TIME,
  ARG_CURRENT_TIME,
  /* FILL ME */
};

GST_PADTEMPLATE_FACTORY (sink_factory,
  "sink",
  GST_PAD_SINK,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW (
    "movdecoder_sink",
    "video/mov",
    NULL
  )
)

GST_PADTEMPLATE_FACTORY (src_video_factory,
  "video_[00-32]",
  GST_PAD_SRC,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW (
    "src_video",
      "video/raw",
      "fourcc",    GST_PROPS_LIST (
                     GST_PROPS_FOURCC (GST_MAKE_FOURCC ('Y','V','1','2')),
                     GST_PROPS_FOURCC (GST_MAKE_FOURCC ('I','Y','U','V')),
                     GST_PROPS_FOURCC (GST_MAKE_FOURCC ('I','4','2','0'))
                   ),
      "width",     GST_PROPS_INT_RANGE (16, 4096),
      "height",    GST_PROPS_INT_RANGE (16, 4096)
  )
)

GST_PADTEMPLATE_FACTORY (src_audio_factory,
  "audio_[00-32]",
  GST_PAD_SRC,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW (
    "src_audio",
      "audio/raw",
      "format",     GST_PROPS_STRING ("int"),
      "law",        GST_PROPS_INT (0),
      "endianness", GST_PROPS_INT (G_BYTE_ORDER),
      "signed",     GST_PROPS_BOOLEAN (TRUE),
      "depth",      GST_PROPS_INT (16),
      "width",      GST_PROPS_INT (16),
      "rate",       GST_PROPS_INT_RANGE (11025, 44100),
      "channels",   GST_PROPS_LIST (
                      GST_PROPS_INT (1),
                      GST_PROPS_INT (2)
                    )
  )
)

static void gst_mov_decoder_class_init(GstMovDecoderClass *klass);
static void gst_mov_decoder_init(GstMovDecoder *mov_decoder);

static void gst_mov_decoder_get_arg(GtkObject *object,GtkArg *arg,guint id);
//static void gst_mov_decoder_set_arg(GtkObject *object,GtkArg *arg,guint id);

static GstPadTemplate *src_video_template, *src_audio_template, *sink_template;

static GstElementClass *parent_class = NULL;
//static guint gst_mov_decoder_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_mov_decoder_get_type(void) {
  static GtkType mov_decoder_type = 0;

  //  printf("gst_mov_decoder_get_type : begin\n");
  if (!mov_decoder_type) {
    static const GtkTypeInfo mov_decoder_info = {
      "GstMovDecoder",
      sizeof(GstMovDecoder),
      sizeof(GstMovDecoderClass),
      (GtkClassInitFunc)gst_mov_decoder_class_init,
      (GtkObjectInitFunc)gst_mov_decoder_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    mov_decoder_type = gtk_type_unique (GST_TYPE_ELEMENT, &mov_decoder_info);
  }
  //  printf("gst_mov_decoder_get_type : end\n");
  return mov_decoder_type;
}

static void
gst_mov_decoder_class_init (GstMovDecoderClass *klass) 
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  gtk_object_add_arg_type("GstMovDecoder::bit_rate", GTK_TYPE_LONG,
                           GTK_ARG_READABLE, ARG_BIT_RATE);
  gtk_object_add_arg_type("GstMovDecoder::media_time", GTK_TYPE_LONG,
                           GTK_ARG_READABLE, ARG_MEDIA_TIME);
  gtk_object_add_arg_type("GstMovDecoder::current_time", GTK_TYPE_LONG,
                           GTK_ARG_READABLE, ARG_CURRENT_TIME);

  parent_class = gtk_type_class (GST_TYPE_ELEMENT);
  
  gtkobject_class->get_arg = gst_mov_decoder_get_arg;

}

static void 
gst_mov_decoder_init (GstMovDecoder *mov_decoder) 
{
  guint i;
		
  //  printf("gst_mov_decoder_init : begin\n");
  mov_decoder->sinkpad = gst_pad_new_from_template (sink_template, "sink");
  gst_element_add_pad (GST_ELEMENT (mov_decoder), mov_decoder->sinkpad);

  gst_element_set_loop_function (GST_ELEMENT (mov_decoder), gst_mov_decoder_loop);

  //GST_FLAG_SET( GST_OBJECT (mov_decoder), GST_ELEMENT_NO_SEEK);

  for(i=0; i<GST_MOVDECODER_MAX_AUDIO_PADS; i++) mov_decoder->audio_pad[i] = NULL;
  for(i=0; i<GST_MOVDECODER_MAX_VIDEO_PADS; i++) mov_decoder->video_pad[i] = NULL;

  //  printf("gst_mov_decoder_init : end\n");
}


static GstCaps*
mov_typefind (GstBuffer *buf,
	      gpointer private)
{
  gchar *data = GST_BUFFER_DATA (buf);
  GstCaps *new;

  new = gst_caps_new ("mov_typefind","video/mov", NULL);

  if (strncmp (&data[4], "wide", 4) == 0) return new;
  if (strncmp (&data[4], "moov", 4) == 0) return new;
  if (strncmp (&data[4], "mdat", 4) == 0) return new;

  return NULL;
}


static void 
gst_mov_decoder_get_arg (GtkObject *object,
		         GtkArg *arg,
		         guint id) 
{
  GstMovDecoder *src;

  g_return_if_fail (GST_IS_MOVDECODER (object));

  src = GST_MOVDECODER (object);

  switch(id) {
    case ARG_BIT_RATE:
      break;
    case ARG_MEDIA_TIME:
      GTK_VALUE_LONG (*arg) = (src->tot_frames*src->time_interval)/1000000;
      break;
    case ARG_CURRENT_TIME:
      GTK_VALUE_LONG (*arg) = (src->current_frame*src->time_interval)/1000000;
      //      printf("Getting CURRENT_TIME %d\n", GTK_VALUE_LONG (*arg));
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}


/*
static GstBuffer *	
gst_mov_decoder_get(GstPad *pad) {
  printf("gst_mov_decoder_get : begin\n");
  printf("gst_mov_decoder_get : end\n");
  return NULL;
}
*/


static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  //  printf("Quicktime plugin_init : begin\n");
  factory = gst_elementfactory_new("quicktime_decoder",GST_TYPE_MOVDECODER,
                                   &gst_mov_decoder_details);
  g_return_val_if_fail(factory != NULL, FALSE);

  src_audio_template = GST_PADTEMPLATE_GET (src_audio_factory);
  gst_elementfactory_add_padtemplate (factory, src_audio_template);

  src_video_template = GST_PADTEMPLATE_GET (src_video_factory);
  gst_elementfactory_add_padtemplate (factory, src_video_template);

  sink_template = GST_PADTEMPLATE_GET (sink_factory);
  gst_elementfactory_add_padtemplate (factory, sink_template);

  gst_plugin_add_type(plugin, &movfactory);

  gst_plugin_add_factory(plugin,factory);

  //  printf("Quicktime plugin_init : end\n");
  return TRUE;
}

GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "quicktime",
  plugin_init
};


static void
gst_mov_decoder_loop (GstElement *element)
{
  GstMovDecoder *mov_decoder;
  quicktime_t *QtFile;
  GstBuffer *buf = gst_buffer_new();
  //MetaVideoRaw meta;


  g_return_if_fail (element != NULL);
  g_return_if_fail (GST_IS_MOVDECODER (element));

  mov_decoder = GST_MOVDECODER (element);

  QtFile = quicktime_open_pad(mov_decoder);
  if(!QtFile)
    printf("gst_mov_decoder_init: QtFile == NULL ");

  mov_decoder->file = QtFile;
  
  /* timestamp stuff */
  mov_decoder->next_time = 0;
  mov_decoder->time_interval =  1000000/quicktime_frame_rate(QtFile, 0);
  mov_decoder->tot_frames = quicktime_video_length(QtFile, 0);
  mov_decoder->current_frame = 0;

  mov_decoder->num_audio_pads = quicktime_audio_tracks(QtFile);
  mov_decoder->num_video_pads = quicktime_video_tracks(QtFile);

  if(mov_decoder->num_video_pads) {
    /* meta stuff */
	  /*
    meta.width = quicktime_video_width(QtFile, 0);
    meta.height = quicktime_video_height(QtFile, 0);
    gst_buffer_add_meta (buf, (GstMeta *)&meta);
    */
    GST_BUFFER_FLAG_SET(buf, GST_BUFFER_DONTFREE);
    // we allocate the data buffer ... very large =)
    GST_BUFFER_SIZE (buf) = 4*quicktime_video_height((quicktime_t*)(mov_decoder->file), 0)*quicktime_video_width((quicktime_t*)(mov_decoder->file), 0);
    GST_BUFFER_DATA (buf) = g_malloc (GST_BUFFER_SIZE (buf));
  } 

  if(mov_decoder->num_audio_pads) {
    if(GST_BUFFER_DATA(buf) == NULL) {
      // Arbitrory size just for the test
      GST_BUFFER_FLAG_SET(buf, GST_BUFFER_DONTFREE);
      GST_BUFFER_SIZE (buf) = 1000000;
      GST_BUFFER_DATA (buf) = g_malloc (GST_BUFFER_SIZE (buf));
    }

  }

  printf("Quicktime header read\n");


  do {
      gst_buffer_ref (buf);
      gst_mov_decoder_chain(mov_decoder->sinkpad, buf);
  } while (!GST_ELEMENT_IS_COTHREAD_STOPPING(element));
}


static void
gst_mov_decoder_chain (GstPad *pad,
		       GstBuffer *buf)
{
  GstMovDecoder *mov_decoder;

  mov_decoder = GST_MOVDECODER (gst_pad_get_parent (pad));

  g_return_if_fail (pad != NULL);
  g_return_if_fail (GST_IS_PAD(pad));
  g_return_if_fail (buf != NULL);
  g_return_if_fail (GST_BUFFER_DATA(buf) != NULL);


  if(mov_decoder->num_video_pads) {
    if(!GST_PAD_CONNECTED(((quicktime_codec_t*)(((quicktime_t*)(mov_decoder->file))->vtracks[0].codec))->priv)) {
      printf("ERROR: OPENQUICKTIME VIDEO NOT CONNECTED\n");
      exit(0);
    } 
    
    if(quicktime_video_position((quicktime_t*)(mov_decoder->file), 0) >= 
       quicktime_video_length((quicktime_t*)(mov_decoder->file), 0)) { 
      // we stop the data source.
      gst_pad_set_eos(mov_decoder->sinkpad);
      return;
    }
    
    // printf("gst_mov_decoder_loop : before quicktime_get_next_video_chunk\n");
    //  buf = quicktime_get_next_video_chunk(mov_decoder->file, 0);
    GST_BUFFER_SIZE(buf) = quicktime_read_frame(mov_decoder->file, GST_BUFFER_DATA (buf), 0);
    GST_BUFFER_TIMESTAMP(buf) = mov_decoder->next_time; 
    mov_decoder->next_time +=  mov_decoder->time_interval;
    mov_decoder->current_frame++;
  }

  if(mov_decoder->num_audio_pads) {
   
  }
  
  if(buf != NULL) {
    gst_pad_push( ((quicktime_codec_t*)(((quicktime_t*)(mov_decoder->file))->vtracks[0].codec))->priv, buf);
  }

  //  gst_buffer_unref (buf);
}

int gst_quicktime_read_data(quicktime_t *file, char *data, longest size)
{
  /* Reading from the GStreamer source pad */
  GstBuffer *buf;

  //  printf("gst_quicktime_read_data : %li bytes\n", size);

  // printf("UTIL reading offset %d%d, size %d\n", file->file_position, size);
  buf = gst_pad_pullregion (((GstMovDecoder*)file->stream)->sinkpad,
			    GST_REGION_OFFSET_LEN,
			    file->file_position, 
			    size);
  if(buf && 
     GST_BUFFER_OFFSET (buf) == file->file_position && 
     GST_BUFFER_SIZE (buf) == size)
    memcpy(data, GST_BUFFER_DATA (buf), GST_BUFFER_SIZE (buf));
  //  else
  //    printf("gstquicktimedecoder: error reading from source\n");

  file->ftell_position += size;
  file->file_position += size;

  return 1;
}


int gst_quicktime_seek(quicktime_t *file, longest offset)
{
  //  printf("gst_quicktime_seek : %li newoffset\n", offset);
  file->ftell_position = offset;
  if(offset > file->total_length || offset < 0) 
    return 1;
  return 0;
}

int gst_quicktime_video_codec(quicktime_t *file, quicktime_video_map_t *vtrack)
{
  int *compressor = (int*)(vtrack->track->mdia.minf.stbl.stsd.table[0].format);
  GstElementFactory *VideoDecfactory;
  GstElement *Decoder;
  GstPad *srcpad, *sinkpad, *newpad;

  //  printf("gst_quicktime_video_codec : begin\n");
  vtrack->codec = (quicktime_codec_t*)calloc(1, sizeof(quicktime_codec_t));

  /* Add new video codecs here */ 
  switch (*compressor) {
  case GST_QUICKTIME_3IV1:
    VideoDecfactory = gst_elementfactory_find("3ivxdec");
    g_return_val_if_fail(VideoDecfactory != NULL, 1);
    
    Decoder = gst_elementfactory_create(VideoDecfactory, "3ivxdec");

    gst_element_set_state(GST_ELEMENT(gst_object_get_parent(GST_OBJECT(file->stream))), GST_STATE_PAUSED);
    gst_bin_add(GST_BIN(gst_object_get_parent(GST_OBJECT(file->stream))), Decoder);

    newpad = gst_pad_new("video", GST_PAD_SRC);
    gst_pad_set_parent(newpad, GST_OBJECT(file->stream));

    sinkpad = gst_element_get_pad(Decoder,"sink");
    srcpad = gst_element_get_pad(Decoder,"src");

    gst_pad_connect(newpad, sinkpad);
    gst_pad_set_name(srcpad, "video_00");
    gst_pad_set_chain_function (newpad, GST_RPAD_CHAINFUNC (sinkpad));

    (*(quicktime_codec_t*)vtrack->codec).priv = newpad;

    gst_element_set_state(GST_ELEMENT(gst_object_get_parent(GST_OBJECT(file->stream))), GST_STATE_PLAYING);

    gtk_signal_emit_by_name (GTK_OBJECT (file->stream), "new_pad", srcpad);

    break;
  default:
    printf("ERROR : currently unsupported video codec %c%c%c%c\n", ((char*)compressor)[0],((char*)compressor)[1],((char*)compressor)[2],((char*)compressor)[3]);
    return 1;
  }

  //  printf("gst_quicktime_video_codec : end\n");

  /* everything is fine now ;) */
  return 0;
}




int gst_quicktime_audio_codec(quicktime_t *file, quicktime_audio_map_t *atrack)
{
  // we don't support audio in this release
  return 0;
}




quicktime_t* quicktime_open_pad(GstMovDecoder *mov_decoder)
{
  quicktime_t *new_file = calloc(1, sizeof(quicktime_t));

  //  printf("quicktime_open_pad : begin\n");
  quicktime_init(new_file);
  new_file->wr = 0;
  new_file->rd = 1;
  new_file->mdat.atom.start = 0;
  new_file->stream = mov_decoder;
  
  new_file->decompressed_buffer_size = 0;
  new_file->decompressed_buffer = NULL;
  new_file->decompressed_position = 0;

  new_file->quicktime_read_data = gst_quicktime_read_data;
  new_file->quicktime_write_data = NULL;
  new_file->quicktime_fseek = gst_quicktime_seek;
  new_file->quicktime_init_vcodec = gst_quicktime_video_codec;
  new_file->quicktime_init_acodec = gst_quicktime_audio_codec;

  if(mov_decoder->sinkpad)
    {
      // Get length.
      new_file->total_length = 1000000000;
      
      if(quicktime_read_info(new_file))
	{
	  //quicktime_close(new_file);
	  fprintf(stderr, "quicktime_open: error in header\n");
	  new_file = 0;
	}
    }

  //  printf("quicktime_open_pad : end\n");
  return new_file;
}
