/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#ifndef __GST_MOVDECODER_H__
#define __GST_MOVDECODER_H__

//#include <quicktime.h>
#include <config.h>
#include <gst/gst.h>



#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#define GST_TYPE_MOVDECODER \
  (gst_mov_decoder_get_type())
#define GST_MOVDECODER(obj) \
  (GTK_CHECK_CAST((obj),GST_TYPE_MOVDECODER,GstMovDecoder))
#define GST_MOVDECODER_CLASS(klass) \
  (GTK_CHECK_CLASS_CAST((klass),GST_TYPE_MOVDECODER,GstMovDecoder))
#define GST_IS_MOVDECODER(obj) \
  (GTK_CHECK_TYPE((obj),GST_TYPE_MOVDECODER))
#define GST_IS_MOVDECODER_CLASS(obj) \
  (GTK_CHECK_CLASS_TYPE((klass),GST_TYPE_MOVDECODER))


typedef enum {
  GST_MOVDECODER_OPEN		= GST_ELEMENT_FLAG_LAST,

  GST_MOVDECODER_FLAG_LAST	= GST_ELEMENT_FLAG_LAST + 2,
} GstMovDecoderFlags;


#define GST_MOVDECODER_MAX_AUDIO_PADS	8	
#define GST_MOVDECODER_MAX_VIDEO_PADS	8	

typedef struct _GstMovDecoder GstMovDecoder;
typedef struct _GstMovDecoderClass GstMovDecoderClass;

struct _GstMovDecoder {
  GstElement element;

  /* pads */
  GstPad *sinkpad;

  /* File Name */
  gchar *filename;

  /* Quicktime structure */
  void *file;

  /* current timestamp */
  guint64 next_time;
  /* time between 2 frames */
  guint64 time_interval;
  /* total number of frames */  
  gulong tot_frames;
  /* current frame */
  gulong current_frame;
 

  guint num_audio_pads;
  guint num_video_pads;
  GstPad       *audio_pad[GST_MOVDECODER_MAX_AUDIO_PADS];
  gboolean 	audio_need_flush[GST_MOVDECODER_MAX_AUDIO_PADS];

  GstPad       *video_pad[GST_MOVDECODER_MAX_VIDEO_PADS];
  gboolean 	video_need_flush[GST_MOVDECODER_MAX_VIDEO_PADS];

  gpointer extra_data;
};

struct _GstMovDecoderClass {
  GstElementClass parent_class;
};

GtkType 		gst_mov_decoder_get_type		(void);
GstElement*		gst_mov_decoder_new			(gchar *name);

GstPad*			gst_mov_decoder_get_video_srcpad	(GstMovDecoder *mov_decoder, 
								 guint pad_nr, GstPadTemplate *temp);
GstPad*			gst_mov_decoder_get_audio_srcpad	(GstMovDecoder *mov_decoder, 
								 guint pad_nr, GstPadTemplate *temp);


/* Quicktime Fourcc */
#define MAKE_FOUR_CC(a,b,c,d) ( ((guint32)a)     | (((guint32)b)<< 8) | \
                                ((guint32)c)<<16 | (((guint32)d)<<24) )
										
#define GST_QUICKTIME_3IV1 MAKE_FOUR_CC('3','I','V','1')


#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* __GST_MOVDECODER_H__ */
