/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "rtpsend.h"
#include "rtp.h"

/* elementfactory information */
static GstElementDetails rtpsend_details = {
  "RTP plugin",
  "RtpSend",
  "Streams data via RTP",
  VERSION,
  "Zaheer Merali <zaheer@bellworldwide.net>",
  "(C) 2001",
};

/* RtpSend signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_IP,
  ARG_PORT,
  ARG_MTU
};

static GstPadTemplate*
sink_factory (void)
{
  return
    gst_padtemplate_new (
  	"sink",
  	GST_PAD_SINK,
 	GST_PAD_ALWAYS,
	NULL);
}

static void			gst_rtpsend_class_init	(GstRtpSendClass *klass);
static void			gst_rtpsend_init		(GstRtpSend *rtpsend);

static void			gst_rtpsend_chain		(GstPad *pad, GstBuffer *buf);

static void			gst_rtpsend_set_arg		(GtkObject *object, GtkArg *arg, guint id);
static void			gst_rtpsend_get_arg		(GtkObject *object, GtkArg *arg, guint id);

static GstElementStateReturn	gst_rtpsend_change_state	(GstElement *element);

GstPadTemplate *src_template, *sink_template;

static GstElementClass *parent_class = NULL;
//static guint gst_icecastsend_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_rtpsend_get_type(void)
{
  static GtkType rtpsend_type = 0;

  if (!rtpsend_type) {
    static const GtkTypeInfo rtpsend_info = {
      "GstRtpSend",
      sizeof(GstRtpSend),
      sizeof(GstRtpSendClass),
      (GtkClassInitFunc)gst_rtpsend_class_init,
      (GtkObjectInitFunc)gst_rtpsend_init,
      (GtkArgSetFunc)gst_rtpsend_set_arg,
      (GtkArgGetFunc)gst_rtpsend_get_arg,
      (GtkClassInitFunc)NULL,
    };
    rtpsend_type = gtk_type_unique(GST_TYPE_ELEMENT,&rtpsend_info);
  }
  return rtpsend_type;
}

static void
gst_rtpsend_class_init (GstRtpSendClass *klass)
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtk_object_add_arg_type("GstRtpSend::ip", GTK_TYPE_STRING,
                          GTK_ARG_READWRITE, ARG_IP);
  gtk_object_add_arg_type("GstRtpSend::port", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_PORT);
  gtk_object_add_arg_type("GstRtpSend::mtu", GTK_TYPE_INT,
			  GTK_ARG_READWRITE, ARG_MTU);

  gtkobject_class->set_arg = gst_rtpsend_set_arg;
  gtkobject_class->get_arg = gst_rtpsend_get_arg;

  gstelement_class->change_state = gst_rtpsend_change_state;
}

static void
gst_rtpsend_init (GstRtpSend *rtpsend)
{
  rtpsend->sinkpad = gst_pad_new_from_template (sink_template, "sink");
  gst_element_add_pad(GST_ELEMENT(rtpsend),rtpsend->sinkpad);
  gst_pad_set_chain_function(rtpsend->sinkpad,gst_rtpsend_chain);

  rtpsend->ip = g_strdup ("127.0.0.1");
  rtpsend->port = 8000;
  rtpsend->mtu = 160;
}

static void
gst_rtpsend_chain (GstPad *pad, GstBuffer *buf)
{
  GstRtpSend *rtpsend;
  //glong ret;
  guint32 i;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);

  rtpsend = GST_RTPSEND (GST_OBJECT_PARENT (pad));

  g_return_if_fail (rtpsend != NULL);
  g_return_if_fail (GST_IS_RTPSEND (rtpsend));

  for(i=0;i<GST_BUFFER_SIZE(buf);i+=rtpsend->mtu) {
    if (GST_BUFFER_SIZE(buf)-i>rtpsend->mtu) {
      rtp_send (&rtpsend->conn, GST_BUFFER_DATA (buf)+i,
		rtpsend->mtu,0,20);
    }
    else {
      rtp_send (&rtpsend->conn, GST_BUFFER_DATA (buf)+i,
		GST_BUFFER_SIZE (buf) - i,0,20);
    }
  }
  /*if (!ret) {
    g_warning ("rtp send error...\n");
    }*/
  
  /*
  shout_sleep (&icecastsend->conn);
  */
  gst_buffer_unref (buf);
}

static void
gst_rtpsend_set_arg (GtkObject *object,GtkArg *arg,guint id)
{
  GstRtpSend *rtpsend;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_RTPSEND(object));
  rtpsend = GST_RTPSEND(object);

  switch(id) {
  case ARG_IP:
    if (rtpsend->ip)
      g_free (rtpsend->ip);
    rtpsend->ip = g_strdup (GTK_VALUE_STRING (*arg));
    break;
  case ARG_PORT:
    rtpsend->port = GTK_VALUE_INT (*arg);
    break;
  case ARG_MTU:
    rtpsend->mtu = GTK_VALUE_INT (*arg);
  default:
    break;
  }
}

static void
gst_rtpsend_get_arg (GtkObject *object,GtkArg *arg,guint id)
{
  GstRtpSend *rtpsend;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_RTPSEND(object));
  rtpsend = GST_RTPSEND(object);

  switch (id) {
  case ARG_IP:
    GTK_VALUE_STRING(*arg) = rtpsend->ip;
    break;
  case ARG_PORT:
    GTK_VALUE_INT(*arg) = rtpsend->port;
    break;
  case ARG_MTU:
    GTK_VALUE_INT(*arg) = rtpsend->mtu;
  default:
    arg->type = GTK_TYPE_INVALID;
    break;
  }
}

static GstElementStateReturn
gst_rtpsend_change_state (GstElement *element)
{
  GstRtpSend *rtpsend;

  g_return_val_if_fail (GST_IS_RTPSEND (element), GST_STATE_FAILURE);

  rtpsend = GST_RTPSEND(element);

  GST_DEBUG (0,"state pending %d\n", GST_STATE_PENDING (element));

  /* if going down into NULL state, close the file if it's open */
  switch (GST_STATE_TRANSITION (element)) {
   case GST_STATE_NULL_TO_READY:
     rtp_client_connection (&rtpsend->conn);

     if (rtp_connection_call(&rtpsend->conn,rtpsend->ip,rtpsend->port)) {
       g_print("initialised RTP successfully...\n");
     }
     else {
       g_warning ("couldn't initialise RTP...\n");
       return GST_STATE_FAILURE;
     }
     break;
   case GST_STATE_READY_TO_NULL:
     /*shout_disconnect (&icecastsend->conn);*/
    rtcp_send(&rtpsend->conn);
    break;
   default:
     break;
  }

  /* if we haven't failed already, give the parent class a chance to ;-) */
  if (GST_ELEMENT_CLASS (parent_class)->change_state)
    return GST_ELEMENT_CLASS (parent_class)->change_state (element);

  return GST_STATE_SUCCESS;
}

static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  factory = gst_elementfactory_new("rtpsend", GST_TYPE_RTPSEND, &rtpsend_details);
  g_return_val_if_fail(factory != NULL, FALSE);

  sink_template = sink_factory ();
  gst_elementfactory_add_padtemplate (factory, sink_template);

  gst_plugin_add_factory (plugin, factory);

  return TRUE;
}

GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "rtpsend",
  plugin_init
};
